/*
 * File:    pfodMAX31855.cpp
 * (c)2014 Forward Computing and Control Pty. Ltd
 * see http://www.pfod.com.au
 *
 *  This library is for the Adafruit Thermocouple Amplifier MAX31855 breakout board
 *  http://http://www.adafruit.com/products/269
 *   and type K therocouple
 *  http://www.adafruit.com/products/270
 *
 *  This library differs from the one provided by Adafruit in that
 * the library does not have any delays and returns the raw reading as an int.
 * This means the Arduino is not locked up waiting for a reading and the integer
 * raw reading can be efficiently averaged before converting to a float for display.
 * 
 * NOTE: No argument constructor uses DI_Pin = A5;  CLK_Pin = A3; CS_Pin = A4; 
 *
 * To start a measurement call 
 * triggerSensor() 
 *
 * triggerSensor can be called every loop as subsequent calls do nothing until a reading has been returned by 
 * getTypeKTemperature();
 * 
 * Calls to getTypeKTemperature() will return pfodMAX31855.NO_DATA (0) until a reading is available
 * When a reading is available getTypeKTemperature() will return either the raw reading (16*Kelvin)
 * or < pfodMAX31855.NO_DATA if error (pfodMAX31855.FAULT_OPEN_CIRCUIT (-1) or pfodMAX31855.FAULT_TO_VCC (-2) or pfodMAX31855.FAULT_TO_GND (-3) )
 *
 * Sample code
#include "pfodMAX31855.h"
#include <elapsedMillis.h>

pfodMAX31855 typeKTemp;
elapsedMillis readingTimer;

void setup() {
  Serial.begin(9600);
  for (int i=5; i>0; i--) {
    delay(1000);
    Serial.print(i);
    Serial.print(" ");
  }
  Serial.println();
  readingTimer = 0;
}

void loop() {
  if (readingTimer > 1000) {
    typeKTemp.triggerSensor();
    readingTimer -= 1000; // reset
  }
  int reading = typeKTemp.getTypeKTemperature();

  if (reading != typeKTemp.NO_DATA) { // have a reading
    Serial.print(millis()/1000.0); // output secs since restart
    Serial.print(F(",Sec,")); // csv comma separator
    if (reading < typeKTemp.NO_DATA) {
      Serial.println(reading);
      Serial.println(F(" ,Invalid Reading, ,Invalid Reading"));
    } else {
      Serial.print(typeKTemp.convertToFahrenheit(reading));
      Serial.print(F(",F,"));
      Serial.print(typeKTemp.convertToCelsius(reading));
      Serial.println(F(",C"));
    }
  }
}
*
*/

#include "pfodMAX31855.h"
/**
 * This constructor uses 
 * DI_Pin = A5, is the input data pin
 * CLK_Pin = A3 is the output clock pin to clock the Sensor's data output
 * CS_Pin = A4 is the chip select pin
 */
pfodMAX31855::pfodMAX31855(void) {
  initMAX31855((byte)A5, (byte)A3, (byte)A4); // sets DI_Pin = A5;  CLK_Pin = A3; CS_Pin = A4; 
}

/**
 * This constructor uses 
 * DI_Pin, is the input data pin
 * CLK_Pin is the output clock pin to clock the Sensor's data output
 * CS_Pin is the chip select pin
 */
pfodMAX31855::pfodMAX31855(byte DI_Pin, byte CLK_Pin, byte CS_Pin) {
  initMAX31855(DI_Pin, CLK_Pin, CS_Pin);
}

void pfodMAX31855::initMAX31855(byte DI_Pin, byte CLK_Pin, byte CS_Pin) {
  DIpin = DI_Pin;
  CLKpin = CLK_Pin;
  CSpin = CS_Pin;
  readState = DATA_FINISHED;
  read_result = NO_DATA;
  clockCounter = 0;
  data = 0;
  clkTimer = 0;
  startupTimer = 0;
  started = false;
  digitalWrite(CSpin, HIGH); // avoid an initial enable
  digitalWrite(CLKpin, HIGH);
  pinMode(CLKpin, OUTPUT);
  pinMode(CSpin, OUTPUT);
  pinMode(DIpin, INPUT);
}

/**
 * Trigger sensor to start taking reading
 * First trigger is delayed by 2sec to allow the MAX38155 to settle
 *
 * You can call this every loop and it will have no effect
 * until getTypeKTemperature returns a non-zero result
 * After getTypeKTemperature returns a non-zero result
 * the next call to triggerSensor will start another reading
 */
void pfodMAX31855::triggerSensor() {
  if (!started && (startupTimer > 500000)) {
    started = true; //  MAX31855 has settled
  }
  if (started && (readState == DATA_FINISHED)) {
    readState = START_READ;
  } // else already reading or waiting for MAX31855 to settle
}

/**
 * Call this each loop to see if reading has been received
 * Returns NO_DATA (0) if still waiting for reading
 * else returns reading in 16* Kelvin.
 * or < NO_DATA if data is invalid for any reason.
 *  INVALID_DATA (-1) if check sum error or
 *  READING_TIMEOUT (-2) if timed out waiting for reading
 *
 * dataType determines reading returned IRTEMP_DATA_IR returns IR temp,
 * any other value returns the ambient temp
 * hard coded in this library to always return IR temp
 */
int pfodMAX31855::getTypeKTemperature(void) {
  if (readState == START_READ) {
    // start new reading
    read_result = NO_DATA;
    readState = SELECT_TO_LOW;
    clockCounter = 32; // counts down to 32 transitions to LOW on clk line
    data = 0;
  } else if (readState == DATA_FINISHED) {
    return NO_DATA;//read_result; // reading available, fault etc
  } // else state == GETTING_DATA

  // else if (state == GETTING_DATA)
  if (readState == SELECT_TO_LOW) {
    digitalWrite(CSpin, LOW); // make select low
    // first data available in 100nS (0.1uS)
    readState = GETTING_DATA;
    clkTimer = 0; // start clock timer for 1uS
  } else if (clkTimer > clkDelay) {
    if (readState == CLK_FINISHED) {
      read_result = processReading(data);
      //  last_result = read_result; // save last result  
      readState = DATA_FINISHED;
      return read_result;
    } // else 
    // do next clock
    int currentClock = digitalRead(CLKpin);
    if (currentClock == HIGH) {
      digitalWrite(CLKpin, LOW);
      // data available in 40nS (0.04uS)
    } else {
      // clock low read data then make high
      // this will be the case 1uS after CS made low
      data <<= 1;
      if (digitalRead(DIpin)) {
        data |= 1;
      }
      clockCounter--;
      if (clockCounter != 0) {
        digitalWrite(CLKpin, HIGH);
      } else {
        readState = CLK_FINISHED;
        digitalWrite(CSpin, HIGH); // make select HIGH to update next reading
      }
    }
    clkTimer = 0; // wait 1uS again
  }
  return NO_DATA;
} // else nothing to do


/*
 processReading returns either 16*degK or one of
 const int FAULT_OPEN_CIRCUIT = -1;
 const int FAULT_TO_VCC = -2;
 const int FAULT_TO_GND = -3;                                       
 */
int pfodMAX31855::processReading(uint32_t reading32) {	
  int errors = reading32 & 0x7;
  if (errors > 0) {
    //typeKReading = INVALID_DATA;
    if (errors & 0x1) {
      return FAULT_OPEN_CIRCUIT;
    } else if (errors & 0x2) {
      return FAULT_TO_VCC;
    } else {
      return FAULT_TO_GND;
    }
  } //else {
  // no errors
  // get the top 14 bits, sign bit is NOT extended
  int16_t reading = reading32 >> 18;
  reading <<= 2; // fixed sign
  reading = reading + kelvinOffset; // extends to size int
  // }
  return reading;
}

/**
 * Convert 16 * Kelvin reading to Centigrade
 */
float pfodMAX31855::convertToCelsius(int reading) {
  if (reading < 0) {
    // invalid 
    return -274.0;
  } // else 
  return (((float)reading-kelvinOffset) / 16.0);
}

/**
 * Convert 16 * Kelvin reading to Fahrenheit
 */
float pfodMAX31855::convertToFahrenheit(int reading) {
  // div by 16 multiply by 9/5
  // then add offset to get fahrenheit
  return (((float)reading-kelvinOffset) * 0.1125 + 32); // -273.15*(9/5) + 32
}


